/*
 * Rewritten from WMix-3.0 (timecop [timecop@japan.co.jp])
 */
#import "Mixer.h"
#import <fcntl.h>

static const char *channel_names[] = SOUND_DEVICE_NAMES;

@implementation Mixer

/* Converts separate left and right channel volumes (each in [0, 1]) to
 * volume and balance values. (Volume is in [0, 1], balance is in [-1, 1])
 */
void lr_to_vb(float left, float right, float *volume, float *balance)
{
    assert((left >= 0.0) && (right >= 0.0));

    *volume = MAX(left, right);

    if (left > right)
        *balance = -1.0 + right / left;
    else if (right > left)
        *balance = 1.0 - left / right;
    else
        *balance = 0.0;
}

/* Performs the reverse calculation of lr_to_vb() */

void vb_to_lr(float volume, float balance, float *left, float *right)
{
    *left = volume * (1.0 - MAX(0.0, balance));
    *right = volume * (1.0 + MIN(0.0, balance));
}

- (void) setState
{
  float left, right;
  int bothVolume, rightVolume, leftVolume;

  if ([mixerChannel[currentChannel] isMuted]) {
    left = 0.0;
    right = 0.0;
  } else {
    vb_to_lr([mixerChannel[currentChannel] volume],
             [mixerChannel[currentChannel] balance],
             &left, &right);
  }
  leftVolume = (int) (100.0 * left);
  rightVolume = (int) (100.0 * right);
  bothVolume = ((rightVolume << 8) | leftVolume);
  ioctl(mixer_fd, MIXER_WRITE([mixerChannel[currentChannel] deviceNumber]),
                              &bothVolume);
}

- (BOOL) updateState
{
  int bothVolume, leftVolume, rightVolume;
  float left, right;
  int srcmask, ch;
  float volume, balance;

  if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &srcmask) == -1) {
      NSLog(@"mixer read failed");
      return NO;
    }

  for (ch = 0; ch < totalChannels; ch++) {
    if (ioctl(mixer_fd, MIXER_READ([mixerChannel[ch] deviceNumber]), 
              &bothVolume) == -1)
      {
         NSLog(@"mixer read failed\n");
         return NO;
      }

    if (bothVolume != [mixerChannel[ch] lastVolume]) {
          leftVolume = bothVolume & 0xFF;
          rightVolume = bothVolume >> 8;

        if ((leftVolume > 0) || (rightVolume > 0))
           [mixerChannel[ch] setIsMuted: NO];

        left = (float) leftVolume / 100.0;
        right = (float) rightVolume / 100.0;

        if (![mixerChannel[ch] isMuted]) {
          if ([mixerChannel[ch] isStereo]) {
             lr_to_vb(left, right, &volume, &balance);
             [mixerChannel[ch] setVolume: volume];
             [mixerChannel[ch] setBalance: balance];
          } else {
             [mixerChannel[ch] setVolume: left];
             [mixerChannel[ch] setBalance: 0.0];
          }

          [mixerChannel[ch] setLastVolume: bothVolume];
        }
     }
     [mixerChannel[ch] setIsRecording: 
                  ((1 << [mixerChannel[ch] deviceNumber]) & srcmask) != 0];

/* For debug
     NSLog(@"--- MIXER ---");
     NSLog(@"Name: %@", [mixerChannel[ch] name]);
     NSLog(@"Device Number: %d", [mixerChannel[ch] deviceNumber]);
     NSLog(@"Last Volume: %d", [mixerChannel[ch] lastVolume]);
     NSLog(@"Volume: %f", [mixerChannel[ch] volume]);
     NSLog(@"Balance: %f", [mixerChannel[ch] balance]);
     NSLog(@"Can Record ? %d", [mixerChannel[ch] canRecord]);
     NSLog(@"Is Recording ? %d", [mixerChannel[ch] isRecording]);
     NSLog(@"Is Stereo ? %d", [mixerChannel[ch] isStereo]);
     NSLog(@"Is Muted ? %d", [mixerChannel[ch] isMuted]);
*/
  }

  return YES;
}

- (BOOL) updateRecordState
{
  int srcmask, ch;

  if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &srcmask) == -1) {
     NSLog(@"mixer read failed");
     return NO;
  }

  for (ch = 0; ch < totalChannels; ch++) {
     [mixerChannel[ch] setIsRecording: 
                  (((1 << [mixerChannel[ch] deviceNumber]) & srcmask) != 0)];
  }
  return YES;  
}

- (BOOL) setRecordState
{
    int srcmask;

    if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &srcmask) == -1) {
      NSLog(@"error: recording source mask ioctl failed");
      return NO;
    }

    if (((1 << [mixerChannel[currentChannel] deviceNumber]) & srcmask) == 0)
        srcmask |= (1 << [mixerChannel[currentChannel] deviceNumber]);
    else
        srcmask &= ~(1 << [mixerChannel[currentChannel] deviceNumber]);

    if (ioctl(mixer_fd, SOUND_MIXER_WRITE_RECSRC, &srcmask) == -1) {
        NSLog(@"error: recording source mask ioctl failed");
        return NO;
    }
    return YES;
}

- initWithDevice: (NSString *) aDevice
{
  int devmask, srcmask, recmask, stmask;
  int count, mask;
 
  self = [super init];

  TEST_RELEASE(device);
  device = aDevice;
  RETAIN(device);

  totalChannels = 0;
  currentChannel = 0;

  mixer_fd = open([device cString], O_RDWR);

  if (mixer_fd == -1) {
      NSLog(@"error: cannot open mixer device %@", device);
      return nil;
    }

  if (ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &devmask) == -1) {
      NSLog(@"error: device mask ioctl failed");
      return nil;
    }

  if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &srcmask) == -1) {
      NSLog(@"error: recording source mask ioctl failed");
      return nil;
    }

  if (ioctl(mixer_fd, SOUND_MIXER_READ_RECMASK, &recmask) == -1) {
      NSLog(@"error: recording mask ioctl failed");
      return nil;
    }

  if (ioctl(mixer_fd, SOUND_MIXER_READ_STEREODEVS, &stmask) == -1) {
      NSLog(@"error: stereo mask ioctl failed");
      return nil;
    }

  for (count = 0; count < SOUND_MIXER_NRDEVICES; count++) 
    {
       mask = 1 << count;
       if (mask & devmask) {
          mixerChannel[totalChannels] = [[MixerChannel alloc] init];
          [mixerChannel[totalChannels] setName: 
                       [NSString stringWithCString: channel_names[count]]];
          [mixerChannel[totalChannels] setDeviceNumber: count];
          [mixerChannel[totalChannels] setLastVolume: -1];
          [mixerChannel[totalChannels] setCanRecord: ((mask & recmask) != 0)];
          [mixerChannel[totalChannels] setIsRecording: ((mask & srcmask) != 0)];
          [mixerChannel[totalChannels] setIsStereo: ((mask & stmask) != 0)];
          [mixerChannel[totalChannels] setIsMuted: NO];
          ++totalChannels;
// Debug    NSLog(@"  %d: %s\n", totalChannels, channel_names[count]);
        }
    }
  if ([self updateState] == NO)
    return NO;

  return self;
}

- (void) dealloc
{
  RELEASE(device);
  [super dealloc];
}

- (void) setCurrentChannel: (int) number
{
  if ((currentChannel >= 0) && (currentChannel < totalChannels)) {
    currentChannel = number;
    [self updateRecordState];
  }
}

- (int) currentChannel
{
  return currentChannel;
}

- (int) totalChannels
{
  return totalChannels;
}

- (NSArray *) allChannelNames
{
  int i; 
  NSMutableArray *array = [NSMutableArray new];

  for(i = 0; i < totalChannels; i++)
    [array addObject: [mixerChannel[i] name]];

  return AUTORELEASE(array);
}

- (NSString *) currentName
{
  return [mixerChannel[currentChannel] name];
}

- (float) currentVolume
{
  [self updateState];
  return [mixerChannel[currentChannel] volume];
}

- (void) setCurrentVolume: (float) number
{
  if((number >= 0.0) && (number <= 1.0))
    {
      [mixerChannel[currentChannel] setVolume: number];
      [self setState];
    }
}

- (float) currentBalance
{
  [self updateState];
  return [mixerChannel[currentChannel] balance];
}

- (void) setCurrentBalance: (float) balance
{
  if((balance >= -1) && (balance <= 1.0)) {
    if ([mixerChannel[currentChannel] isStereo]) {
      [mixerChannel[currentChannel] setBalance: balance];
      [self updateState];
    }
  }
}

- (BOOL) isCurrentMuted
{
  return [mixerChannel[currentChannel] isMuted];
}

- (void) setCurrentMute: (BOOL) aBool
{
  [mixerChannel[currentChannel] setIsMuted: aBool];
  [self setState];
}

- (BOOL) isCurrentRecord
{
  return [mixerChannel[currentChannel] isRecording];
}

- (void) setCurrentRecord: (BOOL) aBool
{
  if ([mixerChannel[currentChannel] canRecord]) {
    [mixerChannel[currentChannel] setIsRecording: aBool];
    [self setRecordState];
    [self updateRecordState];
  }
}

- (BOOL) currentCanRecord
{
  return [mixerChannel[currentChannel] canRecord];
}

- (BOOL) isCurrentStereo
{
  return [mixerChannel[currentChannel] isStereo];
}
@end
